#include <getopt.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <assert.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <limits.h>

#include "input.h"
#include "display.h"
#include "output.h"
#include "boxes.h"
#include "format.h"
#include "opts.h"
#include "reduce.h"
#include "results.h"


void
usage() {
   INFO_ "This is %s version %s\n", PACKAGE_NAME, PACKAGE_VERSION _INFO;
   INFO_ "\nUsage: %s [options] file1 ... fileN\n", PACKAGE_NAME _INFO;
   INFO_ "\nBoxes and options:\n" _INFO;
   INFO_ "    -i        %s box\n", NEEDLE_IDENTITY_BOX->name _INFO;
   INFO_ "    -l        %s box\n", LENGTH_BOX->name _INFO;
#ifndef NO_BACKTRACE
   INFO_ "    -b        Produce pairwise alignment files (also requires -k)\n" _INFO;
#endif
   INFO_ "    -f float  Alignment score threshold\n" _INFO;
   INFO_ "    -m int    Top n results (default 100)\n" _INFO;
   INFO_ "    -p nb     Number of frames explored (3 or 6)\n" _INFO;
   INFO_ "    -a        Produce a posteriori multiple alignment\n" _INFO;
   INFO_ "         -t       Use T-coffee (default is ClustalW)\n" _INFO;
   INFO_ "         -d       Use Dialign2 (default is ClustalW)\n" _INFO;
#ifdef HTML_ALIGNMENT
   INFO_ "         -w       Produce colorized HTML alignment\n" _INFO;
#endif
   INFO_ "    -z        Force Protea's prediction to CODING\n"_INFO;

   INFO_ "\nMiscellaneous options:\n" _INFO;
   INFO_ "    -h        This help and exit\n" _INFO;
   INFO_ "    -q        Quiet (no info messages)\n" _INFO;
   INFO_ "    -k        Keep temporary files\n" _INFO;
   INFO_ "    -r        Use existing pairwise alignment files\n" _INFO;
   /*   INFO_ "    -r float   Similar sequences threshold (default is %.1f)\n", SIM_THRESHOLD _INFO;*/
   INFO_ "    -o dir    Puts generated and kept files in the directory 'dir'\n" _INFO;

   INFO_ "\nBuiltin parameters:\n" _INFO;
#ifdef PROTEA_REDUCE
   INFO_ "    - Reducing dataset before computing alignments (id >= %.2f%%)\n",PROTEA_SIM_THRESHOLD  _INFO;
#endif
#ifndef MY_NEEDLE
   INFO_ "    - Use needle program (EMBOSS package) to produce alignments\n" _INFO;
#else
   INFO_ "    - Use builtin Needleman&Wunsch to produce alignments\n" _INFO;
#ifdef GLOBAL_NEEDLE
   INFO_ "          * Use GLOBAL alignment\n" _INFO;
#else
   INFO_ "          * Use SEMI-GLOBAL alignment\n" _INFO;
#endif
#endif
#ifdef ADAPTIVE_BLOSUM
   INFO_ "          * Use related BLOSUM matrix\n" _INFO;
#else
   INFO_ "          * Use BLOSUM62 matrix only\n" _INFO;
#endif

#ifdef NB_RANK
#ifdef RANKING_ABOVE
   INFO_ "    - Use ranking system above %i edges (best %i pairwise alignments)\n", RANKING_ABOVE, NB_RANK _INFO;
#ifdef NORMAL_DISTRIB
   INFO_ "          * Use full system with gaussian distribution below %i edges\n", RANKING_ABOVE _INFO;
#else
   INFO_ "          * Use full system with zscores below %i edges\n", RANKING_ABOVE _INFO;
#endif
#endif
#else
   INFO_ "    - Use full system with gamma distribution\n" _INFO;
#endif
#ifdef PROTEA_POST_FUSION
   INFO_ "    - Merge RFG vertices if id%% >= %.1f%%\n", PROTEA_SIM_THRESHOLD _INFO;
#ifdef PROTEA_PRE_FUSION
   INFO_ "          * Amino acid sequences are deducted from multiple alignment of merged sequences\n" _INFO;
#endif
#ifdef PROTEA_FUSION_MAX_INDEP_SET
   INFO_ "          * Reduce sequence set before merging vertices\n" _INFO;
#else
   INFO_ "          * Merge chained vertices\n" _INFO;
#endif
#ifdef PROTEA_FUSION_SP
   INFO_ "          * Compute FULL sum-of-pairs score on edges\n" _INFO;
#else
   INFO_ "          * Compute PARTIAL sum-of-pairs score on edges\n" _INFO;
#endif
#else
   INFO_ "    - Do not merge RFG vertices\n" _INFO;   
#endif

#ifdef SYNC_STRAND
   INFO_ "    - Simultaneous sequence strand changes (3^n or 2*3^n assignments)\n" _INFO;
#else
   INFO_ "    - Non simultaneous sequence strand changes (3^n or 6^n assignments)\n" _INFO;
#endif


   INFO_ "\nExternal programs paths:\n" _INFO;
#ifndef MY_NEEDLE
   INFO_ "    needle   = '%s'\n", NEEDLE_EXE _INFO;
#endif
   INFO_ "    clustalw = '%s'\n", CLUSTALW_EXE _INFO;
   INFO_ "    tcoffee  = '%s'\n", TCOFFEE_EXE _INFO;
   INFO_ "    dialign2 = '%s'\n", DIALIGN2_2_EXE _INFO;


   INFO_ "\nThis software is under CeCILL license (http://www.cecill.info/index.en.html)\n"_INFO;

   INFO_ "\nFor questions or bug reports, please contact %s\n", PACKAGE_BUGREPORT _INFO;
}






int
main(int argc, char** argv) {
   blackbox* boxes = NULL;
   sequence* seqs = NULL;
   sequence* seq_t = NULL;
   sequence stemp = NULL;
   sequence** prots;
   char *temp;
   long int tli;
   

   char c;

   int nb_boxes, nb_seqs, t, i, tt, nerror;



   FILE *tf /*, *sample_output = NULL*/;


   if(argc == 1) {
      usage();
      exit(-2); /**** ATTENTION ****/
   }



   init_opts();


   NEW(boxes, blackbox, argc);
   nb_boxes = 0;
   nerror = 0;

   while((c = getopt(argc, argv, ":abdf:hkilm:o:p:r:tqwz")) != EOF) {
      switch(c) {
      case 'a':
	 OPTS_multiple_alignment = 1;
	 break;
#ifndef NO_BACKTRACE
      case 'b':
	 if(OPTS_produce_pairwise_alignements == 0) {
	    OPTS_produce_pairwise_alignements = 1;
	 } else {
	    ERROR_ "option 'b' already set !!\n" _ERROR;
	    exit(1);
	 }
	 break;
#endif
      case 'd':
	 if(OPTS_clustalw_multiple_alignment && (OPTS_dialign_multiple_alignment == 0)) {
	    OPTS_clustalw_multiple_alignment = 0;
	    OPTS_dialign_multiple_alignment = 1;
	 } else {
	    ERROR_ "a posteriori alignment method already set !!\n" _ERROR;
	    exit(1);
	 }
	 break;
      case 'f':
	 if(OPTS_min_aln_score == -INFINITY) {
	    OPTS_min_aln_score = atof(optarg);
	 } else {
	    ERROR_ "similarity threshold already defined !!\n" _ERROR;
	    exit(1);
	 }	 
	 break;
      case 'h':
	 usage();
	 exit(2); /**** ATTENTION ****/
	 break;
      case 'i':
	 boxes[nb_boxes++] = NEEDLE_IDENTITY_BOX;
	 break;
      case 'k':
	 OPTS_keep_files = 1;
	 break;
      case 'l':
	 boxes[nb_boxes++] = LENGTH_BOX;
	 break;
      case 'm':
	 if(OPTS_max_res < 0) {
	    OPTS_max_res = atoi(optarg);
	    if(OPTS_max_res < 0) {
	       ERROR_ "Cannot display less than 1 result\n" _ERROR;
	       exit(1);
	    }
	 } else {
	    ERROR_ "max results already defined !!\n" _ERROR;
	    exit(1);
	 }
	 break;
      case 'o':
	 if(OPTS_output_dir == NULL) {
	    NEW(OPTS_output_dir, char, (strlen(optarg) + 2));
	    strcpy(OPTS_output_dir, optarg);
	    if(OPTS_output_dir[strlen(optarg) - 1] != '/') {
	       OPTS_output_dir[strlen(optarg)] = '/';
	    } else {
	       OPTS_output_dir[strlen(optarg)] = '\0';
	    }
	    OPTS_output_dir[strlen(optarg) + 1] = '\0';
	    
	    if(prepare_dir(OPTS_output_dir))
	       exit(1);
	 } else {
	    ERROR_ "output directory already defined !!\n" _ERROR;
	    exit(2);
	 }
	 break;
      case 'p':
	 if(OPTS_first_phase == 0) {
	    switch(atoi(optarg)) {
	    case 3:
	       OPTS_first_phase = 1;
	       break;
	    case 6:
	       OPTS_first_phase = -3;
	       break;
	    default:
	       ERROR_ "invalid number of frames '%i': it must be 3 or 6\n", atoi(optarg) _ERROR;
	       exit(1);
	    }
	 } else {
	    ERROR_ "number of frames already defined !!\n" _ERROR;
	    exit(1);
	 }
	 break;
      case 'r':
	 if(OPTS_use_cache == 0) {
	    OPTS_use_cache = 1;
	 } else {
	    ERROR_ "option 'r' already set !!\n" _ERROR;
	    exit(1);
	 }
	 break;
      case 't':
	 if(OPTS_clustalw_multiple_alignment && (OPTS_dialign_multiple_alignment == 0)) {
	    OPTS_clustalw_multiple_alignment = 0;
	 } else {
	    ERROR_ "a posteriori alignment method already set !!\n" _ERROR;
	    exit(1);
	 }
	 break;
      case 'q':
	 OPTS_quiet = 1;
	 break;
      case 'w':
#ifdef HTML_ALIGNMENT
	 OPTS_html_alignment = 1;
	 break;
#endif
      case 'z':
	 OPTS_force_protea_prediction = 1;
	 break;
      case ':':
      default:
	 ERROR_ "invalid option '-%c', or missing files !\n", optopt _ERROR;
	 exit(3);
	 break;
      }
   }


   nb_seqs = 0;
   while(optind < argc) {
      
      if(read_sequences(argv[optind], &seq_t, &t)) {
	 ERROR_ "problem while reading %s\n", argv[optind] _ERROR;
	 exit(4);
      } else {
	 for(i = 0; i < t; i++) {
	    check_sequence(seq_t[i]);
	    
	    stemp = ungap(seq_t[i]);
	    DESTROY(seq_t[i]->name);
	    DESTROY(seq_t[i]->input_name);
	    DESTROY(seq_t[i]->file);
	    DESTROY(seq_t[i]->bases);
	    DESTROY(seq_t[i]);
	    
	    seq_t[i] = stemp;
	    
	    if(seq_t[i]->length == 0) {
	       ERROR_ "Empty sequence '%s'\n", seq_t[i]->input_name _ERROR;
	       exit(4);
	    }
	    
	    
	    stemp = traduct(seq_t[i]);
	    DESTROY(seq_t[i]->name);
	    DESTROY(seq_t[i]->input_name);
	    DESTROY(seq_t[i]->file);
	    DESTROY(seq_t[i]->bases);
	    DESTROY(seq_t[i]);
	    
	    seq_t[i] = stemp;
	    NEW(temp, char, (strlen(seq_t[i]->name) + 1 + 10));
	    sprintf(temp, "seq_%i_%s", (nb_seqs + i), seq_t[i]->name);
	    DESTROY(seq_t[i]->name);
	    seq_t[i]->name = temp;
	    
	 }
	 
	 if(nb_seqs) { RENEW(seqs, sequence, (t+nb_seqs)); }
	 else { NEW(seqs, sequence, t); }
	 
	 for(i = 0; i < t; i++) {
	    seqs[nb_seqs + i] = seq_t[i];
	 }
	 /*memcpy(&(seqs[nb_seqs]), seq_t, t*sizeof(sequence));*/
	 
	 nb_seqs += t;
	 DESTROY(seq_t);
      }
      
      optind++;
   }


   boxes[nb_boxes++] = NEEDLE_PROTEIN_BOX;


   /* read stdin */
   if(nb_seqs == 0) {
      if(read_sequences(NULL, &seq_t, &t)) {
	 ERROR_ "problem while reading %s\n", argv[optind] _ERROR;
	 exit(4);
      } else {
	 for(i = 0; i < t; i++) {
	    check_sequence(seq_t[i]);
	    
	    stemp = ungap(seq_t[i]);
	    DESTROY(seq_t[i]->name);
	    DESTROY(seq_t[i]->input_name);
	    DESTROY(seq_t[i]->file);
	    DESTROY(seq_t[i]->bases);
	    DESTROY(seq_t[i]);
	    
	    seq_t[i] = stemp;
	    
	    if(seq_t[i]->length == 0) {
	       ERROR_ "Empty sequence '%s'\n", seq_t[i]->input_name _ERROR;
	       exit(4);
	    }
	    
	    
	    stemp = traduct(seq_t[i]);
	    DESTROY(seq_t[i]->name);
	    DESTROY(seq_t[i]->input_name);
	    DESTROY(seq_t[i]->file);
	    DESTROY(seq_t[i]->bases);
	    DESTROY(seq_t[i]);
	    
	    seq_t[i] = stemp;
	    NEW(temp, char, (strlen(seq_t[i]->name) + 1 + 10));
	    sprintf(temp, "seq_%i_%s", (nb_seqs + i), seq_t[i]->name);
	    DESTROY(seq_t[i]->name);
	    seq_t[i]->name = temp;
	    
	 }
	 
	 if(nb_seqs) { RENEW(seqs, sequence, (t+nb_seqs)); }
	 else { NEW(seqs, sequence, t); }
	 
	 for(i = 0; i < t; i++) {
	    seqs[nb_seqs + i] = seq_t[i];
	 }
	 /*memcpy(&(seqs[nb_seqs]), seq_t, t*sizeof(sequence));*/
	 
	 nb_seqs += t;
	 DESTROY(seq_t);
      }
   }


   if(nb_seqs == 0) {
      ERROR_ "Missing sequences...\n" _ERROR;
      exit(5);
   }

   /*   
   for(t = 0; t < nb_seqs; t++) {
      INFO_ "squence %i file = '%s' name = '%s' length = '%i' bases ='%s'\n",(t+1), seqs[t]->file, seqs[t]->name, seqs[t]->length, seqs[t]->bases _INFO;
	fprintf(DISPLAY_OUT, "%s\n", (to_protein(seqs[t])->bases));
   }
   */


   if(nb_seqs < 2) {
      ERROR_ "Not enough sequences (< 2) !! Exiting...\n" _ERROR;
      exit(6);
   }






   /*   if(output_dir == NULL) output_dir = "./";*/
   if(OPTS_first_phase == 0) OPTS_first_phase = 1;
   if(OPTS_max_res == 0) {
      if(OPTS_first_phase < 0) {
#ifdef SYNC_STRAND
	 tli = 2 * quick_pow(3,nb_seqs);
#else
	 tli = quick_pow(6,nb_seqs);
#endif
      } else {
	 tli = 2 * quick_pow(3,nb_seqs);
      }
      OPTS_max_res = MIN(tli, INT_MAX);
   }
   if(OPTS_max_res < 0) {
      OPTS_max_res = 36;
      INFO_ "falling back to default value %i for max_result\n", OPTS_max_res _INFO;
   }
   /*
     if(coding_sequences > 1.0) {
     ERROR_ "invalid value (> 1.0) for type sequences\n" _ERROR;
     exit(3);
   }
   */


   /*
   if(min_similarity < 0.0) {
      min_similarity = 0.0;
      INFO_ "falling back to default value %.2f for min_similarity\n", min_similarity _INFO;
   }
   */

   if(OPTS_min_aln_score == -INFINITY) {
      OPTS_min_aln_score = 1.0;
      INFO_ "falling back to default value %.2f for min_aln_score\n", OPTS_min_aln_score _INFO;
   }

/*    if(sim_th <= 0.0) { */
/*       sim_th = SIM_THRESHOLD; */
/*       INFO_ "falling back to default value %.1f for sim_th\n", sim_th _INFO;             */
/*    } */


   /*
   if((coding_sequences >= 0.0) && output_dir) {
      init_sampler(coding_sequences);
      NEW(temp, char, (strlen(output_dir) + strlen(SAMPLE_OUT) + 1));
      strncpy(temp, output_dir, strlen(output_dir));
      strncat(temp, SAMPLE_OUT, strlen(SAMPLE_OUT));
      if((sample_output = fopen(temp, "w")) == NULL) {
	 ERROR_ "unable to open '%s' for writing\n", temp _ERROR;
	 exit(4);
      }
      DESTROY(temp);
   }
   */









   /*
   write_sequences(FASTA_FORMAT, DISPLAY_OUT, seqs, nb_seqs);

   for(i = 0; i < nb_seqs; i++)
      for(t = -3; t < 4; t++)
	 if(t != 0) {
	    NEW(seq_t, sequence, 1);
	    *seq_t = change_sequence_strand(seqs[i], t);
	    *seq_t = to_protein(*seq_t);
	    write_sequences(FASTA_FORMAT, DISPLAY_OUT, seq_t, 1);
	    DESTROY(seq_t);
	 }
   */


   for(i = 0; i < nb_seqs; i++) {
      if(seqs[i]->file) {
	 DESTROY(seqs[i]->file);
      }
      
      temp = tmpnam(NULL);
      NEW(seqs[i]->file, char, (strlen(temp) + 1));
      strcpy(seqs[i]->file, temp);
      
      if((tf = fopen(seqs[i]->file, "w"))) {
	 write_fasta(tf, &(seqs[i]), 1, 0);
	 fflush(tf);
	 fclose(tf);
      } else {
	 ERROR_ "unable to open file '%s' for writing\n", seqs[i]->file _ERROR;
	 return 2;
      }
   }

/*    reduce_seq_set(seqs, nb_seqs, &fseqs, &fnb_seqs, sim_th); */
/*    if(fnb_seqs < 2) { */
/*       ERROR_ "Not enough non similar sequences (< 2) !! Exiting...\n" _ERROR; */
/*       exit(7); */
/*    } */
   
/*    DESTROY(seqs); */
/*    seqs = fseqs; */
/*    nb_seqs = fnb_seqs; */


   if(OPTS_output_dir) {
      NEW(temp, char, (strlen(OPTS_output_dir) + strlen(SEQUENCES_FILE) + 1));
      strcpy(temp, OPTS_output_dir);
      strcat(temp, SEQUENCES_FILE);
      if((tf = fopen(temp, "w"))) {
	 write_fasta(tf, seqs, nb_seqs, 0);
	 fflush(tf);
	 fclose(tf);
      } else {
	 ERROR_ "unable to open file '%s' for writing\n", temp _ERROR;
	 return 2;
      }
      DESTROY(temp);
   }

   prots = translate_and_write_sequences(seqs, nb_seqs);


   for(t = 0; t < nb_boxes; t++) {
      if(boxes[t]->fun(seqs, nb_seqs, prots)) {
	 nerror++;
	 WARN_ "box '%s' exited abnormally\n", boxes[t]->name _WARN;
      }
   }


   /*
   if((coding_sequences >= 0.0) && output_dir) {
      if(write_sample(sample_output, the_sample)) {
	 ERROR_ "unable to create write in sample file\n" _ERROR;
      }
      fflush(sample_output);
      fclose(sample_output);
   }
   */

   for(i = 0; i < nb_seqs; i++) {
      for(t = OPTS_first_phase; t < 4; t++) {
	 if(t != 0) {
	    tt = (t > 0) ? (t - 1) : (2 - t);
	    unlink(prots[i][tt]->file);

	    DESTROY(prots[i][tt]->file);
	    DESTROY(prots[i][tt]->input_name);
	    DESTROY(prots[i][tt]->name);
	    DESTROY(prots[i][tt]->bases);
	    DESTROY(prots[i][tt]);
	 }
      }

      unlink(seqs[i]->file);

      DESTROY(seqs[i]->file);
      DESTROY(seqs[i]->input_name);
      DESTROY(seqs[i]->name);
      DESTROY(seqs[i]->bases);
      DESTROY(seqs[i]);

      DESTROY(prots[i]);
   }

   DESTROY(seqs);
   DESTROY(prots);

   if(OPTS_output_dir) {
      DESTROY(OPTS_output_dir);
   }


   destroy_results(nb_seqs);
   
   /*   main_aux(seqs, nb_seqs, boxes, nb_boxes);*/

   return nerror;
}
