/*
Copyright or © or CCopr. INRIA : Arnaud FONTAINE

arnaud.fontaine@lifl.fr

This software is a computer program whose purpose is to predict 
significantly conserved protein coding sequences and/or RNA structure 
on a set of unaligned nucleic sequences.

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#define _GNU_SOURCE
#include <math.h>
#include <mpfr.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>




#include "opts.h"
#include "pvalues.h"
#include "display.h"




mpfr_t**
init_comb(const int max_n, const int keep_n) {   
   int i, j;
   mpfr_t **comb_cache = NULL;
   
   NEW(comb_cache, mpfr_t*, (max_n + 1));
   NEW(comb_cache[0], mpfr_t, 1);
   mpfr_init(comb_cache[0][0]);
   mpfr_set_ui(comb_cache[0][0], 1, GMP_RNDN);
   
   for(i = 1; i <= max_n; i++) {
      NEW(comb_cache[i], mpfr_t, (i+1));
      mpfr_inits(comb_cache[i][0], comb_cache[i][i], NULL);
      mpfr_set_ui(comb_cache[i][0], 1, GMP_RNDN);
      mpfr_set_ui(comb_cache[i][i], 1, GMP_RNDN);
      
      for(j = 1; j < i; j++) {
	 mpfr_init(comb_cache[i][j]);
	 mpfr_add(comb_cache[i][j], comb_cache[i-1][j-1], comb_cache[i-1][j], GMP_RNDN);
      }

      if((i-1) != keep_n) {
	 for(j = 0; j <= (i-1); j++) {
	    if(j != (keep_n - 1)) {
	       mpfr_clear(comb_cache[i-1][j]);
	    }
	 }
      }
   }

   return comb_cache;
}


void
clear_comb(mpfr_t** comb_cache, const int max_n, const int keep_n) {   
   int i, j;
   
   
   for(i = 0; i <= max_n; i++) {
      for(j = 0; j <= i; j++) {
	 if((i == keep_n) || (j == (keep_n - 1))) {
	    mpfr_clear(comb_cache[i][j]);
	 }
      }

      DESTROY(comb_cache[i]);
   }

   DESTROY(comb_cache);
}


mpfr_t*
distrib_sum_des(const int nb_des, const int nbr) {
   mpfr_t* res = NULL;
   mpfr_t comb1, comb2, sixn, som;
   mpfr_t **comb_cache = NULL;

   int score_min, score_max, nsize, i, k, idx;

   score_min = nb_des;
   score_max = nbr * nb_des;
   nsize = score_max - score_min + 1;
   
   NEW(res, mpfr_t, nsize);

   mpfr_inits(comb1, comb2, sixn, som, NULL);

   mpfr_ui_pow_ui(sixn, nbr, nb_des, GMP_RNDN);


   comb_cache = init_comb(score_max, nb_des);


   for(i = score_min; i <= score_max; i++) {
      idx = i - score_min;
      mpfr_init(res[idx]);

      mpfr_set_ui(som, 0, GMP_RNDN);
      
      for(k = 0; k <= (i-nb_des)/nbr; k++) {

	 mpfr_mul(res[idx], comb_cache[nb_des][k], comb_cache[i-nbr*k-1][nb_des-1], GMP_RNDN);

	 if((k & 1) == 1) {
	    mpfr_neg(res[idx], res[idx], GMP_RNDN);
	 }
	 
	 mpfr_add(comb2, som, res[idx], GMP_RNDN);
	 mpfr_swap(comb2, som);
      }

      mpfr_div(res[idx], som, sixn, GMP_RNDN);
   }

   clear_comb(comb_cache, score_max, nb_des);

   mpfr_clears(comb1, comb2, sixn, som, NULL);

   return res;
}



#ifdef NB_RANK

mpfr_t
*pvalues_ranking_scores_sum(const int nb_align) {
   mpfr_t *res = NULL;
   mpfr_t temp;

   int i, nsize;
   int score_min, score_max;
   int nbr = (OPTS_first_phase < 0)?(NB_RANK):(NB_RANK/2);
   

   if(nb_align < 2) {
      
      NEW(res, mpfr_t, nbr);
      
      for(i = 0; i < nbr; i++) {
	 mpfr_init(res[i]);
	 mpfr_set_ld(res[i], ((nbr - i) * 1.0l) / nbr, GMP_RNDN);
      }

   } else {
      
      score_min = nb_align;
      score_max = nb_align * nbr;
      nsize = score_max - score_min + 1;
      
      mpfr_set_default_prec((nb_align * (2.0 + log2(nbr))));


      res = distrib_sum_des(nb_align, nbr);
      
      mpfr_init(temp);
      
      for(i = nsize-2; i >= 0; i--) {
	 mpfr_add(temp, res[i], res[i+1], GMP_RNDN);
	 mpfr_swap(res[i], temp);
      }
   
      mpfr_clear(temp);
      
   }

   return res;
}
 
#endif
 

